* 31.25 baud PEK modem
* With serial interface
* Andrew Senior, G0TJZ. 27-June-1996
* See SB2DSP.DOC (when written!) for more details on this code
*
* Version 1.10 9-Aug-96
*   Reset command ('Q') removed. Host now resets DSP via DTR line
*
* Version 1.20 11-Sep-96
*   Changes to RS232 Tx routine: TBITCNT (bit counter) no longer needed, end of
*   frame is detected by an empty shift register (no stop bit left). The routine
*   now also has a data holding register like a real UART. Comments in RS232 code
*   moved 1 tab stop left to prevent fouling up the assembler!
*   Also discovered that previous versions didn't seem to define TXDRDY, yet still
*   worked!
*
* Version 1.30 12-Sep-96
*   Complete re-write of RS232 TX routine, to use bit-rate clock instead of sample
*   counter, similar to the PSK data tx routine.
*
* Version 1.40 13-Sep-96
*   Complete re-write of RS232 RX routine, to use similar bit-clock system to
*   TX routine.
*   Unused locations 'rshfreg' and 'rbitcnt' removed.
*   Unused constants for async. routines removed.
*
* Version 1.41 13-Sep-96
*   Transmit buffer counter added with increment in main program and
*   decrement in PSK transmit routine.
*
* Version 1.42 13-Sep-96
*   PSK transmit routine determines if buffer is empty or not from
*   the buffer counter, not by comparing AR4 & AR5.
*
* Version 1.43 13-Sep-96
*   Urgent holding register added to RS232 TX routine. This register
*   is read checked before the normal holding register when sending
*   a new character.
*
* Version 1.44 13-Sep-96
*   Corrected undetected bug in previous modification: the TXUDRDY flag
*   would never get cleared and TXRDY would get cleared instead.
*
* Version 1.45 13-Sep-96
*   Main prog. now sends 0x07 to PC when buffer is nearly full.
*   LSB of phase data now forced to zero to allow discrimination
*   between it and buffer status data.
*   PSK Tx routine sends 0x03 to PC when buffer has emptied.
*
* Version 1.46 13-Sep-96
*   Corrected bug in above. Mask value for phase data was given
*   as #0FE instead of #0FEh.
*
* Version 1.47 13-Sep-96
*   File send doesn't work. Seems that buffer full signal gets there too late,
*   so buffer full threshold drastically reduced to test this.
*   Tests showed this had no effect. It seems the problem was with the driver
*   program. See SLOWBP2.C
*
* Version 1.50 27-Oct-96
*   Modified to use new 4-bit receive filter designed by G3PLX
*
* Version 1.51 2-Nov-96
*   Receive filter changed to updated design by G3PLX
*
* Version 1.60 25-Nov-96
*   12500/500 Hz decimator frequency control modified to give exact
*   divide-by-25.
*   Transmit reconstruction filter simplified to 25-tap rectangular.
*
* Version 1.61 28-Nov-96
*   Code modified to use PM=1 and BSAR instead of repeated SFRs
*   Local echo added (suggested by PA0IHD)
*
* Version 1.62 13-Dec-96
*   Local echo modified to use on/off set instead of toggle
*
* Version 1.63 5-Jan-97
*   Bug in datime routine which cause tx lock up if incoming data
*   word had bit 15 set. Corrected: bug due to SXM=1 when loading
*   word from buffer into accu for shifting.
*
* Version 1.70 9-Mar-97
*   Modified so that the on/off keying control (by changing the D/A
*   output mask) is delayed so that change occurs at the envelope zero
*   during a phase-reversal. This allows smoother keying for Morse ID.
*         
* Version 1.80 27-Jun-97
*   Modified to send actual signal phase values at each bit centre to
*   host instead of phase difference since last bit centre.
*
* Version 1.90 29-Jun-97
*   Buffer empty notification can be switched on/off by host (Default 
*   is on)

	 .mmregs

TA      .set    16              ; SCAF clock = 312.5 kHz
RA      .set    16
TAp     .set    31
RAp     .set    31
TB      .set    25              ; Fs = 12.5 KHz
RB      .set    25
AIC_CMD .set    19h
;AIC_CMD .set    1Dh             ; Use AUX in

; Data locations
	

ifil1st .set    0100h   ; 1st I-channel LPF (100 taps)
ifil1en .set    0163h
itemp   .set    0164h

ifil2st .set    0165h   ; 2nd I-channel LPF (64 taps)
ifil2en .set    01A4h

qfil1st .set    01A6h   ; 1st Q-channel LPF (100 taps)
qfil1en .set    0209h
qtemp   .set    020Ah

qfil2st .set    020Bh   ; 2nd Q-channel LPF (64 taps)
qfil2en .set    024Ah

tfil2st .set    024Ch   ; Transmit LPF (25 taps)
tfil2en .set    0264h

	.ds     0300h
* PEK demodulator data
rxphase .word   300             ; Receive oscillator phase
rxfreq  .word   5243            ; Oscillator frequency = 1000 Hz
; MSB indicates subtract, remainder is offset to add/subtract
actab   .word   0               ;  0.0 + arctan
	.word   0A000h          ;  4000h - arctan
	.word   08000h          ;  0.0 - arctan
	.word   06000h          ;  C000h + arctan
	.word   0C000h          ;  1.0 - arctan
	.word   02000h          ;  0.5 + arctan
	.word   04000h          ; -1.0 + arctan
	.word   0E000h          ; -0.5 - arctan
inphase .word   0               ; inphase component
quadrat .word   0               ; quadrature component
octant  .word   0               ; octant for arctan
arctan  .word   0               ; partial arctan result
temp    .word   0               ; temporary storage
atcof1  .word   2B20h ; arctan linear coeff
atcof2  .word   0B20h ; arctan squarelaw coeff
saphase .word   25              ; 16x datarate clock
bitphas .word   0               ; Rx clock recovery phase
oldbtph .word   0               ; Clock phase last sample
nowphas .word   0               ; Signal phase this sample
oldphas .word   0               ; Signal phase last sample
echo    .word   0               ; Local echo mask

* PEK modulator data
txphase .word   0               ; Transmit oscillator phase
txfreq  .word   5243            ; Frequency = 1000 Hz
txsig   .word   0               ; Tx output signal buffer
daphase .word   0               ; Datarate clock (31.25 Hz)
newchar .word   32768           ; New character in shift register flag
txbit   .word   32768           ; Bit to transmit
txamptr .word   64              ; Tx amplitude lookup pointer, init for cos lookup
txampl  .word   0               ; Tx amplitude after 2nd filter
txinv   .word   0               ; Modulation sense invert flag
tbufcnt .word   0               ; Transmit buffer counter
oodlyst .space  144             ; Output on/off control delay line
oodlyen .word   0
	.word   0               ; DMOV protection
ntfybuf .word   1               ; Buffer empty notification on/off flag

* Asynchronous comms data
TXDRDY  .WORD   0               ; Tx data ready flag
TXHREG  .WORD   0               ; Tx data holding register
TXUDRDY .WORD   0               ; Tx urgent data ready flag
TXUHREG .WORD   0               ; Tx urgent holding register
TXREG   .WORD   0               ; Tx data (shift) register
ATBTPHS .WORD   0               ; Tx bit clock phase
ARSTATE .WORD   0               ; Rx routine state
ARBTPHS .WORD   0               ; Rx bit clock phase
BITCNT  .WORD   0               ; Rx bit counter
RXREG   .WORD   0               ; Rx data register
RXBUF   .WORD   0               ; Rx data buffer
RXDRDY  .WORD   0               ; Rx data ready flag

rxtemp  .word   0               ; Temp storage in main prog

; Asynchronous comms routine states
SPCWT   .SET    1
SPCCK   .SET    2
RCVDAT  .SET    3

; Lookup tables, etc.
sintab  .set    1000h           ; 256-point sine lookup table
tabend  .set    10FFh           ; end of table
lpfcof1 .set    1100h           ; 100 I/Q low-pass filter coefficients (1st LPF)
lpfcof2 .set    1164h           ; 64 I/Q low-pass filter coefficients (2nd LPF)
txbufst .set    1500h           ; Buffer for data to send via modem
txbufen .set    18FFh           ; End of buffer

	.ds     1000h
	.listoff
	.include "sintab.inc"
	.include "sb100.inc"    ; 100-tap filter
	.include "sb64-2.inc"   ; 64-tap filter
	.liston

* Set up the ISR vectors

	.ps     080ah
rint    B       receive       ;0A; Serial port receive interrupt RINT.
xint    B       transmit      ;0C; Serial port transmit interrupt XINT.

* TMS320C5X INITIALIZATION 

	.ps 0a00h
	.entry
START:  setc    INTM                    ; Disable interrupts
	ldp     #0                      ; Set data page pointer
	opl     #0834h,PMST
	lacc    #0                      ; Zero wait states
	samm    CWSR
	samm    IOWSR
	samm    PDWSR
	clrc    CNF                     ; Map DARAM B0 into data space

	call    SET_AIC                 ; initialize AIC and enable interrupts
	 
	clrc    OVM                     ; OVM = 0
	spm     1                       ; PM = 1
	setc    SXM                     ; Enable sign extension
	splk    #010h,IMR               ; Activate RINT only

; Zero transmit filter buffer

	mar     *,AR1
	lar     AR1,#tfil2st
	lacc    #0
	rpt     #24
	sacl    *+

; Setup buffer pointers

	lar     AR4,#txbufst            ; Tx buffer input pointer
	lar     AR5,#txbufst            ; Tx buffer output pointer

	splk    #txbufst,CBSR1          ; Use CB1 to control AR4
	splk    #txbufen,CBER1
	splk    #txbufst,CBSR2          ; Use CB2 to control AR5
	splk    #txbufen,CBER2
	splk    #0DCh,CBCR
	
; Configure asynchronous comms

	SETC    XF                      ; Mark state
	LDP     #rxphase
	SPLK    #0,TXDRDY               ; No data to send
	SPLK    #SPCWT,ARSTATE          ; Waiting for start bit
	SPLK    #0,RXDRDY               ; No data received yet

	mar     *,AR4           ; AR4 is tx buf in pointer
	clrc    INTM
dawait  call    getbyte         ; Get byte from PC
	sacl    rxtemp
	cpl     #70,rxtemp      ; If 'F', set rx freq
	bcnd    setfreq,TC
	cpl     #78,rxtemp      ; If 'N', net tx freq to rx freq
	bcnd    nettx,TC
	cpl     #68,rxtemp      ; If 'D', transfer data to tx buffer
	bcnd    dodata,TC
	cpl     #69,rxtemp      ; If 'E', toggle local echo mode
	bcnd    setecho,TC
	cpl     #66,rxtemp      ; If 'B', set buffer empty notify mode
	bcnd    setntfy,TC
	b       dawait

setfreq call    getbyte         ; Get MS byte of data
	sacl    rxtemp          ; Store it
	call    getbyte         ; Get LS byte
	add     rxtemp,8        ; Merge LS and MS bytes
	sacl    rxfreq          ; Update freq regsiter
	b       dawait

nettx   lacc    rxfreq          ; Copy rx freq to tx freq register
	sacl    txfreq
	b       dawait

dodata  call    getbyte         ; Get MS byte of data
	sacl    rxtemp          ; Store it
	call    getbyte         ; Get LS byte
	add     rxtemp,8        ; Merge LS and MS bytes
	sacl    *+              ; Store in buffer (AR4 controlled by CB1)
	lacl    tbufcnt         ; Get buffer count
	add     #1              ; Increment
	sacl    tbufcnt         ; Store new count
	sub     #1000,0         ; Check if buffer is nearly full
	bcnd    dawait,LEQ      ; If <= 1000 characters, return
	splk    #7,TXUHREG      ; Send buffer full signal
	splk    #1,TXUDRDY
	b       dawait

setecho call    getbyte         ; Get echo mode byte
	bcnd    echooff,EQ
	splk    #0FFFFh,echo    ; Set echo mask 'on'
	b       dawait
echooff splk    #0,echo         ; Set echo mask 'off'
	b       dawait

setntfy call    getbyte         ; Get notify mode byte
	sacl    ntfybuf         ; Set flag
	b       dawait

getbyte bit     RXDRDY,15       ; Wait for data ready flag
	bcnd    getbyte,NTC
	lacc    RXBUF           ; Load byte into accu
	splk    #0,RXDRDY       ; Clear flag
	ret     
	
;------- end of main program ----------;

receive mar     *,AR0
	ldp     #rxphase
	lamm    DRR             ; Get latest sample
	sacb                    ; Store in AccB
	lacl    txsig           ; Get tx output
	and     echo            ; Mask with local echo mask
	addb                    ; Add A/D sample
	samm    TREG0           ; Store in T
	lacc    rxphase         ; Get current osc phase
	add     rxfreq          ; Update it
	sacl    rxphase         ; Store updated phase
	bsar    8               ; Extract MS byte
	and     #00FFh
	sacl    temp            ; Save for later
	add     #sintab,0       ; Acc = sin lookup address
	samm    AR0             ; AR0 = sin address
	lacc    temp
	add     #64             ; Offset for cos
	and     #00FFh
	add     #sintab,0       ; Acc = cos lookup address
	samm    AR1             ; AR1 = cos address
	mpy     *,AR1           ; P = T * sin(wt)
	ldp     #ifil1st
	sph     ifil1st         ; Store in 1st I filter
	mpy     *               ; P = T * cos(wt)
	ldp     #qfil1st
	sph     qfil1st         ; Store in 1st Q filter
; Now do 1st I channel filter
	zap
	lar     AR1,#ifil1en    ; AR1 -> last tap
	rpt     #99             ; 100 taps
	macd    lpfcof1,*-      ; Do the filter
	apac
	ldp     #ifil2st
	sach    ifil2st         ; Store I result in 2nd filter
; Now do 1st Q channel filter
	zap
	lar     AR1,#qfil1en
	rpt     #99
	macd    lpfcof1,*-
	apac
	ldp     #qfil2st
	sach    qfil2st        ; Store Q result in 2nd filter

* Generate output waveform
; Run transmit reconstruction filter
	ldp     #rxphase
	lacc    txampl          ; Get running total
	ldp     #tfil2st
	sub     tfil2en         ; Subtract oldest sample
	add     tfil2st         ; Add newest sample
	ldp     #rxphase
	sacl    txampl          ; Store new total
	lar     AR1,#tfil2en
	rpt     #24
	dmov    *-              ; Shift delay line
; Now form output PEK signal
	ldp     #rxphase
	lacc    txphase         ; Get current osc phase
	add     txfreq          ; Update it
	sacl    txphase         ; Store updated phase
	bsar    8               ; Extract MS byte
	and     #00FFh
	add     #sintab,0       ; Accu = sine lookup address
	tblr    txsig           ; Fetch the waveform value to txsig
	lt      txsig           ; T = tx sine wave
	mpy     txampl          ; P = tx amplitude * tx sine wave
	sph     txsig
	lacc    txsig           ; Get into low accu
	and     oodlyen         ; Mask value before sending
	sacl    txsig           ; Store for local echo
	samm    DXR             ; Output it
	
; See if it's time to run 500 Hz processes
	ldp     #rxphase
	lacc    saphase         ; Get counter value
	sub     #1              ; Decrement
	sacl    saphase         ; Store new count
	bcnd    rxdone,NEQ      ; Branch if non-zero
	lacl    #25             ; Re-init counter
	sacl    saphase
	call    satime          ; Do 500 Hz processes

; Run asynchronous comms routines
rxdone  call    ASYTX
	call    ASYRX
	rete

* Start of 500 Hz (16x datarate) processes

; Now do 2nd I channel filter
satime  zap
	lar     AR1,#ifil2en
	rpt     #63             ; 64 taps
	macd    lpfcof2,*-
	apac
	ldp     #inphase
	sach    inphase         ; Store result for arctan
; Now do 2nd Q channel filter
	zap
	lar     AR1,#qfil2en
	rpt     #63
	macd    lpfcof2,*-
	apac
	ldp     #quadrat
	sach    quadrat         ; Store result for arctan

* Now do arctan routine
* This code adapted from 320C26 code written by G4BMK
* from algorithm by G3PLX.

	ldp     #actab
	setc    SXM
; Now compute the OCTANT in AR1 (current Aux.Reg.)
	lar     AR1,#0          ; Octant = zero
	lacc    inphase         ; test inphase
	bcnd    oct2,GEQ
	neg                     ; Negate accumulator
	sacl    inphase         ; - restore in 'inphase'
	adrk    #4              ; Add 4 to octant
oct2    lacc    quadrat
	bcnd    oct4,GEQ
	neg
	sacl    quadrat         ; negate quadrature
	adrk    #2              ; add 2 to octant
oct4    sub     inphase         ; acc = quadrature - inphase
	bcnd    oct6,LEQ        ; skip if inphase >= quadrature
	adrk    #1              ; add 1 to octant
	lacl    inphase         ; swap inphase<>quadrat
	add     quadrat,16
	sach    inphase
	sacl    quadrat
oct6    sar     AR1,octant      ; save resultant octant

; Now divide quadrature by inphase
; At this stage they are both positive values and quadrature < inphase
; Fractional division. See C2x manual, Pg 5-59
	lacc    quadrat,16      ; Get numerator into Acc-Hi
	rpt     #14             ; (15-bit shift)
	subc    inphase         ; conditional subtract of denominator
	sacl    temp            ; Acc-lo contains quotient (tangent)
	samm    TREG0           ; tangent into T as well
	lacc    #0              ; clear Acc
	mpy     temp            ; P = sqr(tan)
	mpya    atcof1          ; P = atcof1*tan, Acc = sqr(tan)
	sach    temp            ; temp = sqr(tan)
	ltp     temp            ; T = sqr(tan), Acc = atcof1*tan
	mpy     atcof2          ; P = atcof2*sqr(tan)
	spac                    ; Acc = atcof1*tan - atcof2*sqr(tan)
	sach    arctan          ; Store arctan
; Pick up an action word from table indexed by octant
	lacc    #actab          ; Form address of entry
	add     octant
	samm    BMAR
	nop                     ; Pipeline protection
	bldd    BMAR,temp       ; Read table entry into temp
	lacc    temp,16         ; load it into Acc-Hi
	sfl                     ; shift sign bit into carry
	bcnd    atan2,C         ; branch if 'twas -ve
	add     arctan,16
	b       atan4
atan2   sub     arctan,16
atan4   sach    nowphas         ; Save full phase angle
	lacc    bitphas         ; Update bit clock
	add     #4096,0
	sacl    bitphas
	lacc    nowphas         ; Get current phase
	sub     oldphas         ; Subtract phase last sample
	abs
	sub     #16384,0        ; Test if change > 90 deg
	bcnd    nocorr,LEQ      ; Don't correct bit clk if <= 90 deg
	lt      bitphas
	mpy     #31785          ; bitphas = bitphas * 0.97
	sph     bitphas
nocorr  lacc    oldbtph         ; Get bit phase last sample
	sub     bitphas         ; Subtract current phase
	bcnd    notctr,LEQ      ; If bitphas >= oldbtph then not bit centre
	clrc    SXM
	lacc    nowphas,8
	and     #0FEh,16        ; Mask LSB of phase data
	sach    TXHREG          ; Store in RS232 tx reg
	setc    SXM
	splk    #1,TXDRDY       ; Set data ready flag
notctr  lacc    bitphas
	sacl    oldbtph         ; Update old bit phase
	lacc    nowphas
	sacl    oldphas         ; Update old signal phase               
; Generate Tx amplitude value
	lacc    txamptr         ; Get amplitude pointer
	bit     txbit,0         ; Test state of bit to send
	bcnd    nochng,NTC      ; Branch if zero
	add     #8              ; Increment pointer if we're sending a 1
	and     #0FFh           ; Modulo 256
	sacl    txamptr         ; Save new value
nochng  add     #sintab,0       ; Form cosine lookup address
	ldp     #tfil2st        ; Read value into start of 2nd tx filter
	tblr    tfil2st
	lt      tfil2st         ; Get value
	mpy     #1310           ; Multiply by 1/25 for rectangular filter
	sph     tfil2st         ; Store result in buffer
	ldp     #rxphase
; Shift on/off control delay line
	lar     AR1,#oodlyen    ; Point to end of line
	rpt     #9
	dmov    *-
; See if it's data clock time
	clrc    C
	lacc    daphase,16      ; Get dataclock phase
	add     #8192,15        ; Increment
	sach    daphase         ; Save updated phase
	bcnd    datime,C        ; Branch if carry
sadone  ret

* Datarate processes

datime  lacl    tbufcnt         ; Get tx buffer count
	bcnd    shift,NEQ       ; Branch if non-zero, i.e. buffer not empty
	splk    #8000h,txbit    ; Force 1 while idle
	b       sadone
shift   mar     *,AR5           ; AR5 is tx buffer pointer
	clrc    SXM             ; Don't sign extend shift reg.
	lacc    *               ; Get current shift register contents
	bit     newchar,0       ; Test if this is a newly arrived character
	bcnd    notnew,NTC      ; Branch if not
	splk    #0,newchar      ; Clear new status
	sfr                     ; Shift out tx invert status into C
	bcnd    setinv,C        ; Branch if status=on
	splk    #0,txinv        ; Clear invert flag
	b       chkoos
setinv  splk    #8000h,txinv    ; Set invert flag
chkoos  sfr                     ; Shift out on/off status bit into C
	bcnd    seton,C         ; Branch if status=on
	splk    #0,oodlyst      ; Mask off all output to D/A
	b       notnew
seton   splk    #0FFFCh,oodlyst ; Set normal D/A mask   
notnew  sfr                     ; Shift LSB into C
	setc    SXM
	sacl    *               ; Save updated register
	bcnd    bithi,C         ; Branch if LSB was 1
	splk    #0,txbit        ; Send a 0
	b       shfchk
bithi   splk    #8000h,txbit    ; Send a 1
shfchk  sub     #1              ; Test if accu = 0001
	bcnd    invchk,NEQ      ; Branch if not
	splk    #8000h,newchar  ; Set new char flag
	mar     *+              ; Point to next location (AR5 ctrl'd by CB2)
	lacl    tbufcnt         ; Get tx buffer count
	sub     #1              ; Decrement
	sacl    tbufcnt
	bcnd    invchk,GT       ; If not zero, continue
	bit     ntfybuf,15      ; Do we need to tell host buffer is empty?
	bcnd    invchk,NTC      ; Branch if not
	splk    #3,TXUHREG      ; Send buffer empty code to PC
	splk    #1,TXUDRDY
invchk  bit     txinv,0         ; Test invert status
	bcnd    dadone,NTC
	lacc    txbit           ; If invert on, invert txbit
	xor     #8000h
	sacl    txbit
dadone  b       sadone


transmit
	rete

*** Asynchronous communications routines ***

*
* Asynchronous comms TX routine
*
ASYTX   CLRC    C
	LACC    ATBTPHS,16      ; Get bit clock phase
	ADD     #16384,15       ; Increment
	SACH    ATBTPHS         ; Save new phase
	BCND    ASYTX_1,C       ; If bit time, call routine
	RET

ASYTX_1 LACC    TXREG           ; Get shift reg.
	BCND    ASYTX_2,NEQ     ; Branch if non-zero
	BIT     TXUDRDY,#15     ; Test if urgent data ready to send
	BCND    ASYTX_4,NTC     ; Branch if not
	LACC    TXUHREG         ; Get word from urgent holding reg.
	SPLK    #0,TXUDRDY      ; Clear flag (holding reg. empty)
	B       ASYTX_5         ; Start sending
ASYTX_4 BIT     TXDRDY,#15      ; Test if data ready to send
	BCND    ASYTX_3,NTC     ; Keep at mark and return if not
	LACC    TXHREG          ; Get word from holding reg.
	SPLK    #0,TXDRDY       ; Clear flag (holding reg. empty)
ASYTX_5 OR      #100h           ; Add stop bit
	SFL                     ; Add start bit
ASYTX_2 SFR                     ; Shift out LSB into C
	SACL    TXREG           ; Store shift reg.
	BCND    ASYTX_3,C       ; Branch if LSB was 1
	CLRC    XF              ; Send space
	RET                     ; Done
ASYTX_3 SETC    XF              ; Send mark
	RET                     ; Done


*
* Asynchronous comms rx routine
*
ASYRX   CPL     #SPCWT,ARSTATE
	BCND    SPC_WT,TC
	CLRC    C
	LACC    ARBTPHS,16      ; Get bit clock phase
	ADD     #16384,15       ; Increment
	SACH    ARBTPHS         ; Store new phase
	BCND    ARBTTIM,C       ; Branch if bit time
	RET

* Called to check for a possible start bit
SPC_WT  BCND    ASYRX_1,BIO     ; Test line state
	RET                     ; Return if mark
ASYRX_1 SPLK    #32768,ARBTPHS  ; Re-init bit clock
	SPLK    #SPCCK,ARSTATE  ; Go to space check state
	RET

* Called at bit centres
ARBTTIM CPL     #SPCCK,ARSTATE
	BCND    SPC_CK,TC
	CPL     #RCVDAT,ARSTATE
	BCND    RCV_DAT,TC
	RET

* Called to check if start-bit is real
SPC_CK  BCND    ASYRX_2,BIO     ; Test line state
	SPLK    #SPCWT,ARSTATE  ; If mark, return to waiting
	RET
ASYRX_2 SPLK    #0,ARBTPHS      ; Re-init bit clock
	SPLK    #RCVDAT,ARSTATE ; Start receiving data
	SPLK    #9,BITCNT       ; Init bit counter
	SPLK    #0,RXREG        ; Zero shift reg.
	RET

* Called to receive a byte
RCV_DAT LACL    RXREG           ; Get shift register
	CLRC    SXM             ; Don't sign extend on shift
	SFR                     ; Shift regisiter right
	SETC    SXM
	BCND    ASYRX_3,BIO     ; Test line state
	OR      #100h           ; If mark, insert a '1'
ASYRX_3 SACL    RXREG           ; Store updated reg.
	LACL    BITCNT          ; Get bit counter
	SUB     #1              ; Decrement
	SACL    BITCNT          ; Store new count
	BCND    ASYRX_4,EQ      ; Branch if zero
	RET
ASYRX_4 LACL    RXREG           ; Get shift reg.
	AND     #0FFh           ; Mask off stop bit
	SACL    RXBUF           ; Store rx'd byte
	SPLK    #1,RXDRDY       ; Set data ready flag
	SPLK    #SPCWT,ARSTATE  ; Wait for next start bit
	RET             
	
* Initialise the AIC

SET_AIC splk    #1,PRD          ; divide by 2 to give 10 MHz on TOUT
	splk    #20h,TCR        ; Reset the timer
	splk    #08h,SPC        ; Set FSM and reset serial port
	splk    #0C8h,SPC       ; Set FSM and take port out of reset
	lacc    #80h
	sacl    GREG            ; Config. 8000h-FFFFh as global mem
	mar     *,AR0
	lar     AR0,#0FFFFh
	rpt     #10000          ; Keep BR low for 0.5ms to reset AIC
	lacc    *,0,AR0
	sach    GREG            ; Cancel global memory
	splk    #20h,IMR        ; Unmask XINT only
	clrc    INTM            ; Enable interrupts
	clrc    SXM
	lacc    #0              ; Setup TA and RA registers
	add     #RA,2
	add     #TA,9
	call    aic2nd
	lacc    #1              ; Setup TA' and RA' registers
	add     #RAp,2          
	add     #TAp,9
	call    aic2nd
	lacc    #2              ; Setup TB and RB registers
	add     #RB,2
	add     #TB,9
	call    aic2nd
	lacc    #3              ; Setup control register
	add     #AIC_CMD,2
	call    aic2nd
	setc    SXM
	setc    INTM            ; Mask interrupts
	ret

aic2nd  sach    DXR
	idle
	add     #6,15           ; Dummy to get 2ndary comms cycle
	sach    DXR             ; Send dummy
	idle                    ; Wait until finished sending
	sacl    DXR             ; Send command
	idle                    ; Wait again
	lacc    #0
	sacl    DXR
	idle
	ret

	.end

