/*
 * Driver program for SLOWBP2.ASM for the TMS320C50 DSK
 * (C) Andrew Senior, G0TJZ, 1996
 * Version 2.32 26-Aug-1997
 *  Supports setting of text and text background colours in SLOWBP2.INI
 * Version 2.31 29-Jun-1997
 *  Modified to turn DSP tx buffer empty notification on or off.
 *  Introduced to prevent a bug which appeared as a result of the
 *  local serial rx buffer.
 *  AFC updates no longer sent when PTT is on. (They didn't do anything
 *  useful anyway!)
 * Version 2.30 27-Jun-1997
 *  Major changes:
 *  Modified to accept actual phase data from DSK instead
 *  of phase differences, as per DSP code version 1.80.
 *  AFC modified to work with the above.
 *  Serial receive buffering added to avoid receive overruns when large
 *  bursts of serial data are sent to the DSK, e.g. on/off control,
 *  CW ident.
 *  File send code move back to main() - it now works as a result of
 *  the buffering.
 * Version 2.20 11-May-1997
 *  AFC implemented.
 * Version 2.10 14-Apr-1997
 *  The levels of RTS and DTR can now be inverted.
 *  A colour display mode is available (colour currently fixed).
 *  Both these are set as options in SLOWBP2.INI along with an option
 *  to configure local echo on/off.
 * Version 2.00 12-Apr-1997
 *  DSK loader incorporated into program.
 * Version 1.91 9-Mar-1997
 *  Modified to support "smooth" on/off keying for Morse ID, etc.
 *  (see SLOWBP2.ASM)
 * Version 1.90 5-Jan-1997
 *  G3PLX DCD scheme added. Tx always sends 32 reversals at start
 *  and 32 bits of carrier at end. Rx switches DCD on after 16
 *  consecutive revs. and off after 16 bits of carrier.
 * Version 1.81 5-Jan-1997
 *  Window "banners" altered to use graphics characters present in
 *  both codepages 437 and 850 for compatibility.
 * Version 1.80 13-Dec-1996
 *  Tx fix/track replaces netting function
 * Version 1.71 13-Dec-1996
 *  Local echo modified to use on/off set instead of toggle.
 *  Symbolic constants added for status display positions.
 * Version 1.70 28-Nov-1996
 *  Local echo (loopback) feature added
 *  Version number display added
 * Version 1.61 28-Sep-1996
 *  Typed characters are not sent if the DSP's transmit buffer is full,
 *  like with file send.
 * Version 1.60 20-Sep-1996
 *  File capture to fixed name file "CAPTURE.TXT" added.
 * Version 1.51 17-Sep-1996
 *  Cosmetic changes to window "banners"
 *  File send prompt line is cleared before & after entering file name.
 * Version 1.50 17-Sep-1996
 *  File send now prompts for a file name.
 * Version 1.41 17-Sep-1996
 *  senddata() function added to tidy up sending data commands
 *  to the DSP.
 * Version 1.40 16-Sep-1996
 *  PTT control via RTS line added.
 * Version 1.33 16-Sep-1996
 *  Program no longer resets DSP at exit time.
 * Version 1.32 16-Sep-1996
 *  Buffer and file send status indicators modified.
 * Version 1.31 13-Sep-1996
 *  File send did not work: the DSP buffer was overrun. Reducing
 *  the threshold in the DSP to make the full signal occurr sooner
 *  had no effect. The problem was resolved by moving the file send
 *  control code from main() to receive(). Obviously something to do
 *  with the timing of the comms.
 * Version 1.30 13-Sep-1996
 *  File send added, from fixed named file, "SLOWBP2.SND".
 * Version 1.22 13-Sep-1996
 *   Fixed bug in previous: was rejecting those with LSB=0 instead!
 *   Now shows DSP buffer status.
 * Version 1.21 13-Sep-1996
 *   In preparation for DSP buffer status detection, receive()
 *   ignores data from the DSP which don't have the LSB=0
 * Version 1.20 9-Aug-1996
 *   Centre frequency and tuning tone frequency now configurable
 *   by a FREQ= option in SLOWBP2.INI.
 *   DSP is now reset by pulsing the DTR line. Not sure if this is any
 *   better than the old method of sending a reset command.
 * Version 1.10 22-Jul-1996
 *   Cosmetic changes: F-key line format improved, Varicode error
 *   symbols changed. Alpha indicates too-long codeword, beta indicates
 *   invalid codeword.
 * Version 1.00 9-Jul-1996
 *   Bug reported by G3YTQ: sending a pound sign (or other non-ASCII code)
 *   causes the DSP send routine to lock up when using Varicode.
 *   Corrected by trapping codes above 127 in sendchar(): the DSP locks
 *   up because it is sent invalid data if the code is > 127.
 * Beta Version 28-Jun-1996
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <conio.h>
#include <time.h>

#define TRUE -1
#define FALSE !TRUE

/* DSK loader commands */
#define LD 2
#define LP 3
#define XG 5

/* loaddsk() error codes */
#define NOFILE 1
#define BADFILE 2
#define NOENTRY 3
#define RESETFAIL 4
#define COMFAIL 5

/* Status display locations */
#define RXF_COL 5
#define RXF_ROW 23
#define TXF_COL 5
#define TXF_ROW 24

#define TRK_COL 24
#define TRK_ROW TXF_ROW

#define TX_COL 30
#define TX_ROW 23
#define CODE_COL 30
#define CODE_ROW 24
#define FIL_COL 30
#define FIL_ROW 25

#define PTT_COL 50
#define PTT_ROW 23
#define CAP_COL 50
#define CAP_ROW 24
#define BUF_COL 50
#define BUF_ROW 25

#define ECHO_COL 65
#define ECHO_ROW 23
#define DCD_COL 65
#define DCD_ROW 24

struct point {
	int x;
	int y;
};

struct winctrl {
	int tlx; /* X co-ord of top left corner */
	int tly; /* Y co-ord of top left corner */
	int brx; /* X co-ord of bottom right corner */
	int bry; /* Y co-ord of bottom right corner */
	int x; /* Current X co-ord */
	int y; /* Current Y co-ord */
};

struct winctrl rxwin = {1, 1, 80, 18, 1, 1};
struct winctrl txwin = {1, 20, 80, 21, 1, 1};
char txwinbanner[] = "͹Transmitted data͹F1:Tx trackF2:Tx on/offF3:Morse IDF4:CodeF5:Sound";
char stwinbanner[] = "͹Tuning and status͹F6:File sendF7:CaptureF8:Echo on/offF9:Toggle DCD";
int combases[4] = {0x3F8, 0x2F8, 0x3E8, 0x2E8};
int base; /* COM port base address */
struct text_info oldmode;
int quit = 0; /* Quit flag */
unsigned int tunepos = 0; /* Tuning display plot position */
struct point tunetab[8] = {{1, 0}, {0, 0}, {0, 1}, {0, 2},
													 {1, 2}, {2, 2}, {2, 1}, {2, 0}};
int bitcount = 0; /* Bit counter for ASCII decoder */
unsigned int rxreg; /* Rx shift register */
int autolf = 1; /* Rx auto-linefeed flag */
double freq = 1000; /* Rx centre frequency */
double txfreq = 1000; /* Tx centre frequency */
double tonefreq = 1000; /* Tuning tone frequency */
int txon; /* Tx on flag */
int vlcode; /* Varicode on flag */
int vlerror = 0; /* Varicode decoder error flag */
int encotab[128], decotab[512]; /* Varicode lookup tables */
char cwidmsg[40] = "NO ID SET"; /* Morse ID message */
int soundon = FALSE; /* Tuning tone on flag */
int dsp_buf_full = FALSE; /* DSP tx buffer full flag */
int file_send = FALSE; /* File send flag */
FILE *snd_fp; /* File send stream pointer */
int ptton = FALSE; /* Status of the PTT line (RTS) */
int dsp_buf_empty = TRUE; /* DSP tx buffer empty flag */
													/* Not just the inverse of dsp_buf_full !! */
int capture = FALSE; /* Capture flag */
FILE *cap_fp; /* File capture stream pointer */
int echo = FALSE; /* Local echo flag */
int txtrack = FALSE; /* Tx tracking flag */
char version[] = "SLOWBP2.EXE Version 2.32 26-Aug-1997";
int dcd = FALSE; /* DCD flag */
int dtrinvert = FALSE; /* DTR invert flag */
int rtsinvert = FALSE; /* RTS invert flag */
int usecolour = TRUE; /* Use colour display flag */
unsigned int loadbaud = 57600; /* DSK loader baudrate */
char *loaderrs[6] = {"No error",
										 "Load file does not exist",
										 "Unrecognised code in DSK file",
										 "No entry point specified in DSK file",
										 "Couldn't reset and initialise DSK",
										 "Communications failure with DSK"};
#define RXBUFSIZE 256
unsigned char rxbuf[RXBUFSIZE]; /* Buffer for received data from DSK */
int rxbufinptr = 0, rxbufoutptr = 0, bufcount = 0; /* Buffer control vars */
int fgcol = YELLOW, bgcol = BLUE; /* Text colour controls */

/* Morse code generation table
 * Covers ASCII 33-90 (! to Z)
 */

unsigned int cwtab[] = {0x14, 0x4A, 0x00, 0x00, 0x00, 0x00, 0x7A, 0xB4,
												0xB6, 0x00, 0x54, 0xCE, 0x86, 0x56, 0x94, 0xFC,
												0x7C, 0x3C, 0x1C, 0x0C, 0x04, 0x84, 0xC4, 0xE4,
												0xF4, 0xE2, 0x00, 0x00, 0x00, 0x00, 0x32, 0x00,
												0x60, 0x88, 0xA8, 0x90, 0x40, 0x28, 0xD0, 0x08,
												0x20, 0x78, 0xB0, 0x48, 0xE0, 0xA0, 0xF0, 0x68,
												0xD8, 0x50, 0x10, 0xC0, 0x30, 0x18, 0x70, 0x98,
												0xB8, 0xC8};

static void initcom(unsigned int baudrate, int nstop)
{
	unsigned char ctl;

	if(nstop == 2)
		ctl = 0x07; /* 2 stop bits */
	else
		ctl = 0x03; /* 1 stop bit */
	outportb(base+3, ctl | 0x80); /* 8,N,nstop access divider */
	outport(base, 115200 / baudrate); /* Set baudrate */
	outportb(base+3, ctl); /* 8,N,nstop access data */
	outportb(base+1, 0x00); /* No interrupts */
	ctl = 0;
	if(!dtrinvert)
		ctl |= 0x01; /* DTR on */
	if(rtsinvert)
		ctl |= 0x02; /* RTS on */
	outportb(base+4, ctl); /* Set control lines */
	inportb(base+5); /* Read spurious indications... */
	inportb(base);
}

/* Insert received data in buffer */

static void getrxdata(void)
{
	rxbuf[rxbufinptr] = inportb(base);
	rxbufinptr = (rxbufinptr + 1) % RXBUFSIZE;
	if(++bufcount > RXBUFSIZE-1) {
		printf("Internal buffer overrun!!!\n");
		exit(1);
	}
}

/* Send byte to DSK */

static void sendbyte(unsigned char u)
{
	unsigned char lsr;

	for(;;) {
		lsr = inportb(base+5);
		if(lsr & 0x01)
			getrxdata();
		if(lsr & 0x20)
			break;
	}
	outportb(base, u);
}

/* Send a byte to DSK and wait for echo in loader mode */

static int sendbytewait(unsigned char byte)
{
	unsigned char u;

	delay(1); /* Ensure DSK ready to receive */
	sendbyte(byte); /* Send it */
	while(!(inportb(base + 5) & 0x01))
		; /* Wait for data ready */
	u = inportb(base); /* Get the echoed byte */
	if(u != byte)
		return 0; /* Bytes don't match */
	return 1; /* OK */
}

/* Send word to DSK and wait for echo */

static int sendwordwait(unsigned int word)
{
	if(!sendbytewait(word / 256)) /* Send MS byte */
		return 0; /* Failed */
	if(!sendbytewait(word % 256)) /* Send LS byte */
		return 0; /* Failed */
	return 1; /* Success */
}

/* Reset DSK and set loader baud rate */

static int resetdsk(void)
{
	int tries;

	outportb(base + 4, inportb(base + 4) ^ 0x01); /* Reset DSK */
	delay(50); /* For 50 msec */
	outportb(base + 4, inportb(base + 4) ^ 0x01); /* Release reset */
	delay(200); /* Wait for DSK to boot */
	inportb(base); /* Read any rubbish */
	for(tries = 10; tries > 0; tries--) {
		sendbyte(0x80); /* Send baudrate setting byte */
		delay(5); /* Allow for DSK to respond */
		if(inportb(base) == 0x1B)
			break; /* ESC received = success */
	}
	if(tries <= 0)
		return 0; /* Failed to get response */
	return 1; /* OK */
}

/* Send load command */

static int startload(int space, unsigned int addr, unsigned int len)
{
	if(!sendbytewait(space)) /* Specify load type */
		return 0; /* Comms failure */
	if(!sendwordwait(addr)) /* Specify start address */
		return 0;
	if(!sendwordwait(len - 1)) /* Specify number of words */
		return 0;
	return 1; /* OK */
}

/* Load object file into DSK and run it */

int loaddsk(char *fname, int baudrate)
{
	FILE *fp;
	char line[128], buf[6], type;
	int done, ptr, count, space;
	unsigned int entry, start, data[32], words = 0;

	initcom(baudrate, 2);
	printf("Resetting DSK... ");
	if(resetdsk())
		printf("OK\n");
	else {
		printf("Failed\n");
		return RESETFAIL;
	}

	fp = fopen(fname, "rt");
	if(fp == NULL)
		return NOFILE; /* Can't open object file */

	printf("Loading...\r");
	for(done = 0; !done;) {
		fgets(line, 128, fp);
		switch(line[0]) {
			case 0:
			case ':':
				done = 1; /* Finished loading */
				break;
			case '1':
				sscanf(&line[1], "%04x", &entry);	/* Entry point */
				break;
			case '9': /* Load data */
				switch(line[5]) {
					case 'M':
						space = LD; /* Data memory */
						break;
					case 'B':
						space = LP; /* Load program */
						break;
					case '7':
						space = 0; /* No data */
						break;
					default:
						fclose(fp);
						return BADFILE; /* Invalid type */
				}
				if(space != 0) {
					sscanf(&line[1], "%04x", &start); /* Start address of pgm or data */
					/* Read data words into array */
					count = 0;
					for(ptr = 5; ;ptr += 5) {
						sscanf(&line[ptr], "%c%04x", &type, &data[count]);
						if(type != '7')
							count++;
						else
							break; /* End of data */
					}
					if(!startload(space, start, count)) { /* Send load command */
						fclose(fp);
						return COMFAIL;
					}
					for(ptr = 0; ptr < count; ptr++) {
						if(!sendwordwait(data[ptr])) { /* Send data values */
							fclose(fp);
							return COMFAIL;
						}
					}
					words += count; /* Total word count */
					printf("Loading... %05d\r", words);
				}
				break;
			default: /* Ignore other lines in file */
				break;
		}
	}
	fclose(fp);
	/* Initialise IMR */
	if(!startload(LD, 4, 1))
		return COMFAIL;
	if(!sendwordwait(2)) /* Unmask INT2 */
		return COMFAIL;
	/* Start the program */
	if(entry != 0) {
		if(!sendbytewait(XG))
			return COMFAIL;
		if(!sendwordwait(entry))
			return COMFAIL;
		sendbyte(0); /* Start pulse - no reply from DSK */
	}
	else
		return NOENTRY; /* No entry point defined */
	printf("Loading... Done. \n");
	return 0;
}

static void init_tables(void)
{
	FILE *fp;
	int k, n;
	unsigned int codeword, mask;
	char buf[12];

	printf("Loading code tables...\n");
	fp = fopen("SLOWBP2.COD", "rt");
	if(fp == NULL) {
		printf("Can't open code table file SLOWBP2.COD\n");
		exit(EXIT_FAILURE);
	}
	for(k = 0; k < 512; k++)
		decotab[k] = 0xE1;	/* Beta indicates invalid codeword */
	for(k = 0; k < 128; k++) {
		codeword = 4;	/* Preset 100 pattern */
		fgets(buf, 12, fp);
		for(n = strlen(buf) - 2; n >= 0; n--) {
			codeword <<= 1;
			if(buf[n] == '1')
				codeword++;
		}
		encotab[k] = codeword;
		mask = 0x8000;	/* Start at MSB */
		while((mask & codeword) == 0)
			mask >>= 1;	/* Find dummy '1' MSB added earlier */
		codeword &= ~mask;	/* Remove it */
		codeword >>= 1;	/* Remove (always '1') LSB */
		decotab[codeword] = k;
	}
	fclose(fp);
}

static void showfreq(void)
{
	gotoxy(RXF_COL, RXF_ROW);
	cprintf("Rx freq= %6.1f Hz", freq);
	gotoxy(TXF_COL, TXF_ROW);
	cprintf("Tx freq= %6.1f Hz", txfreq);
	gotoxy(TRK_COL, TRK_ROW);
	if(txtrack)
		cprintf("Track");
	else
		cprintf("Fix  ");
}

static void displayinit(void)
{
	gettextinfo(&oldmode); /* Read current mode settings */
	if(usecolour) {
		textmode(C80); /* 80-col colour */
		textcolor(fgcol);
		textbackground(bgcol);
		clrscr();
	}
	else
		textmode(BW80); /* Select 80-col black & white */
	_setcursortype(_NOCURSOR);
	gotoxy(2, 24);
	putch('T');
	gotoxy(1, 19);
	cputs(txwinbanner);
	gotoxy(1, 22);
	cputs(stwinbanner);
  showfreq();
}

static void restoredisplay(void)
{
	textmode(oldmode.currmode);
	textattr(oldmode.attribute);
	_setcursortype(_NORMALCURSOR);
	clrscr();
}

static void dispchar(char c, struct winctrl *wcp)
{
	if((wcp == &rxwin) && !dcd)
		return; /* Don't print rx'd data if DCD off */
	window(wcp->tlx, wcp->tly, wcp->brx, wcp->bry);
	gotoxy(wcp->x, wcp->y);
	switch(c) {
		case 8:
			putch(8); /* Backspace */
			putch(' '); /* Erase old character */
			putch(8); /* Backspace again */
			break;
		case 7: /* Ignore bells */
			break;
		case 13:
			if(autolf)
				putch(10);
			/* Note the fall-through */
		default:
			putch(c);
	}
	wcp->x = wherex();
	wcp->y = wherey();
	window(1, 1, 80, 25); /* Restore full-screen window */
	if(capture && wcp == &rxwin)
		fputc(c, cap_fp); /* Capture to file */
}

static void vl_decode(int level)
{
	level ^= 0x80;	/* Invert received level */
	rxreg >>= 1;
	if((rxreg & 0x0008) > 0)
		vlerror = 1;
	if(level)
		rxreg |= 0x8000;
	if(((rxreg & 0xC000) == 0) && (rxreg != 0)) {
		while((rxreg & 0x0001) == 0)
			rxreg >>= 1;
		rxreg >>= 1;
		if(vlerror)
			dispchar(0xE0, &rxwin); /* Display alpha for too-long codeword */
		else
			dispchar(decotab[rxreg], &rxwin);
		vlerror = 0;
		rxreg = 0;
	}
}

static void ascii_decode(int bit)
{
	switch(bitcount) {
		case 0: /* Waiting for start bit */
			if(bit)
				break; /* Still mark */
			bitcount = 1;
			rxreg = 0;
			break;
		case 8: /* Reached stop bit */
			dispchar(rxreg, &rxwin);
			bitcount = 0;
			break;
		default:
			bitcount++;
			rxreg |= bit;
			rxreg >>= 1;
			break;
	}
}

/* Set buffer empty notification on or off */

static void setbuffernotify(int mode)
{
	sendbyte('B');
	if(mode)
		sendbyte(0xFF);
	else
		sendbyte(0x00);
}

static void read_dsp_status(int status)
{
	if(status & 0x02) {
		/* Buffer status information */
		gotoxy(BUF_COL, BUF_ROW);
		if(status & 0x04) {
			dsp_buf_full = TRUE;
      setbuffernotify(1);
			cprintf("BUFFER FULL!");
		}
		else {
			dsp_buf_full = FALSE;
			dsp_buf_empty = TRUE;
			setbuffernotify(0);
			cprintf("            ");
		}
	}
}

static void show_file_send(void)
{
	gotoxy(FIL_COL, FIL_ROW);
	if(file_send)
		cprintf("FILE SEND");
	else
		cprintf("         ");
}

static void do_file_send(void)
{
	char buf[62], *name;

	file_send = !file_send;
	if(file_send) {
		gotoxy(1, 21);
    cputs("                                                            ");
		gotoxy(1, 21);
		cputs("File to send?: ");
		buf[0] = 60;
		name = cgets(buf);
		gotoxy(1, 21);
		cputs("                                                            ");
		snd_fp = fopen(name, "rb");
		if(snd_fp == NULL)
			file_send = FALSE;
	}
	else
		fclose(snd_fp);
	show_file_send();
}

static void senddata(unsigned int u)
{
	dsp_buf_empty = FALSE;
	sendbyte('D');
	sendbyte(u / 256); /* Send MS byte */
	sendbyte(u % 256); /* Send LS byte */
}

static void send_ascii_char(int c)
{
	unsigned int u;

	u = (c << 3) | 0x0C02;
	senddata(u);
}

static void send_vl_char(int c)
{
	unsigned int u;

	u = (encotab[c] << 2) | 0x0003;
	senddata(u);
}

static void sendchar(int c)
{
	if(!txon)
		return; /* Don't try to send if tx off */
	if(c > 127)
		return; /* Can't send codes above 127 */
	if(vlcode)
		send_vl_char(c);
	else
		send_ascii_char(c);
	dispchar(c, &txwin);
}

static void setfreq(double deltaf)
{
	unsigned int w;

	freq += deltaf;
	w = (unsigned int)(freq / 12500.0 * 65536 + 0.5);
	sendbyte('F'); /* Set rx freq command */
	sendbyte(w / 256); /* MS byte */
	sendbyte(w % 256); /* LS byte */
	if(txtrack) {
		txfreq = freq;
		sendbyte('N'); /* Net tx to rx freq */
	}
	showfreq();
}

static void toggletxtrack(void)
{
	txtrack = !txtrack;
	if(txtrack) {
		txfreq = freq;
		sendbyte('N'); /* Net tx freq to rx freq */
	}
	showfreq();
}

static void showtxstatus(void)
{
	gotoxy(TX_COL, TX_ROW);
	if(txon)
		cprintf("TRANSMITTING");
	else
		cprintf("            ");
}

static void showptt(void)
{
	gotoxy(PTT_COL, PTT_ROW);
	if(ptton)
		cprintf("PTT ON");
	else
		cprintf("      ");
}

static void ptt_on(void)
{
	unsigned char ctl;

	ptton = TRUE;
	ctl = inportb(base+4);
	if(rtsinvert)
		ctl &= 0xFD; /* RTS off */
	else
		ctl |= 0x02; /* RTS on */
	outportb(base+4, ctl); /* Turn PTT on */
	showptt();
}

static void ptt_off(void)
{
	unsigned char ctl;

	ptton = FALSE;
	ctl = inportb(base+4);
	if(rtsinvert)
		ctl |= 0x02; /* RTS on */
	else
		ctl &= 0xFD; /* RTS off */
	outportb(base+4, ctl); /* Turn PTT off */
	showptt();
}

/* Send 32 phase reversals at start of over */

static void send_reversals(void)
{
	if(vlcode) {
		senddata(0x8003); /* 13 revs */
		senddata(0x8003); /* + 13 revs */
		senddata(0x0103); /* + 6 revs = 32 */
	}
	else {
		senddata(0xFFFE); /* As above... */
		senddata(0xFFFE);
		senddata(0x01FE);
	}
}

/* Send continuous tone for 32 bits then turn off */

static void send_carrier(void)
{
	senddata(0xFFFF);
	senddata(0xFFFF);
	senddata(0x01FF);
	senddata(0x0009);
}

static void toggletx(void)
{
	txon = !txon;
	if(txon) {
		ptt_on();
		send_reversals();
	}
	else {
		send_carrier();
		setbuffernotify(1);
	}
	showtxstatus();
}

static void send_cw_char(int c)
{
	unsigned char u;

	c = toupper(c);
	if(c < ' ' || c > 'Z')
		return;
	if(c == ' ')
		/* 6 dots space */
		senddata(0x8001);
	else {
		u = cwtab[c - 33];
		while(u > 0) {
			if(u == 0x80)
				/* 2 dots space */
				senddata(0x0041);
			else if(u & 0x80)
				/* 3 dots mark */
				senddata(0x01FB);
			else
				/* 1 dot mark */
				senddata(0x001B);
			/* 1 dot inter-element space */
			senddata(0x0011);
			u = u << 1;
		}
	}
}

static void cwident(char *msg)
{
	int i;

	/* Send +CW+ in PSK before doing ident */
	sendchar('\r');
	sendchar('+');
	sendchar('C');
	sendchar('W');
	sendchar('+');
	sendchar('\r');
	/* Send end-of-transmission tone */
	send_carrier();
	/* Do ident */
	send_cw_char(' ');
	for(i = 0; i < strlen(msg); i++)
		send_cw_char(msg[i]);
	setbuffernotify(1);
	txon = FALSE;
	showtxstatus();
}

static void togglecode(void)
{
	vlcode = !vlcode;
	gotoxy(CODE_COL, CODE_ROW);
	if(vlcode)
		cprintf("VARICODE");
	else
		cprintf("ASCII   ");
	rxreg = 0;
	vlerror = 0;
}

static void setecho(unsigned char mode)
{
	sendbyte('E');
	sendbyte(mode);
}

static void toggle_echo(void)
{
	echo = !echo;
	gotoxy(ECHO_COL, ECHO_ROW);
	if(echo) {
		setecho(0xFF);
		cprintf("ECHO ON");
	}
	else {
		setecho(0x00);
		cprintf("       ");
	}
}

static void togglesound(void)
{
	soundon = !soundon;
	if(soundon)
		sound((unsigned)tonefreq);
	else
		nosound();
}

static void show_capture(void)
{
	gotoxy(CAP_COL, CAP_ROW);
	if(capture)
		cprintf("CAPTURE");
	else
		cprintf("       ");
}

static void toggle_capture(void)
{
	capture = !capture;
	if(capture) {
		cap_fp = fopen("CAPTURE.TXT", "a+b");
		if(cap_fp == NULL)
			capture = FALSE;
	}
	else
		fclose(cap_fp);
	show_capture();
}

static void showdcd(void)
{
	gotoxy(DCD_COL, DCD_ROW);
	if(dcd)
		cprintf("DCD ON ");
	else
		cprintf("DCD OFF");
}

static void dodcd(int bit)
{
	static unsigned int dcdshfreg;

	dcdshfreg <<= 1; /* Shift register left */
	if(bit)
		dcdshfreg |= 1; /* Insert '1' bit */
	if(dcdshfreg == 0xFFFF) {
		dcd = TRUE;
		showdcd();
	}
	else if(dcdshfreg == 0x0000) {
		dcd = FALSE;
		showdcd();
	}
}

static void receive(void)
{
	int offs, nowphase;
	unsigned char lsr;
	static time_t lastsec = 0, nowsec;
	static int lastphase = 0;
	unsigned int diffphase;

	nowphase = rxbuf[rxbufoutptr];
	rxbufoutptr = (rxbufoutptr + 1) % RXBUFSIZE;
	if(bufcount-- < 0) {
		printf("Internal buffer underrun!\n");
		exit(1);
	}
	if(nowphase & 0x01) {
		read_dsp_status(nowphase);
		return;
	}
	diffphase = nowphase - lastphase;
	lastphase = nowphase;
	offs = (char)((diffphase << 1) & 0xFF);
	if(offs == -128)
		offs = 0;
	if(!ptton && dcd)
		freq += offs * 0.0001; /* AFC adjustment */
	time(&nowsec);
	if((nowsec != lastsec) && !ptton)
		setfreq(0); /* Do AFC update */
	lastsec = nowsec;
	gotoxy(tunetab[tunepos].x + 1, tunetab[tunepos].y + 23);
	putch(' ');
	tunepos = (diffphase + 16) % 256 / 32;
	gotoxy(tunetab[tunepos].x + 1, tunetab[tunepos].y + 23);
	putch('*');
	diffphase = (diffphase + 0x40) & 0x80; /* 0x80 => '1', 0x00 => '0' */
	dodcd(diffphase);
	if(vlcode)
		vl_decode(diffphase);
	else
		ascii_decode(diffphase);
}

static void setport(void)
{
	if(base < 1 || base > 4) {
		printf("Port number must be between 1 and 4\n");
		exit(EXIT_FAILURE);
	}
	base = combases[base - 1];
}

static void setcode(char c)
{

	/* NB: The settings of vlcode are "wrong" because they
	 * get inverted during the rest of the initialisation
	 */
	if(c == 'V')
		vlcode = FALSE;
	else if(c == 'A')
		vlcode = TRUE;
	else {
		printf("%c is not a valid setting for CODE\n", c);
		exit(EXIT_FAILURE);
	}
}

static void checkrtsinvert(void)
{
	if(rtsinvert < 0 || rtsinvert > 1) {
		printf("RTSINVERT must be 0 or 1\n");
		exit(EXIT_FAILURE);
	}
}

static void checkdtrinvert(void)
{
	if(dtrinvert < 0 || dtrinvert > 1) {
		printf("DTRINVERT must be 0 or 1\n");
		exit(EXIT_FAILURE);
	}
}

static void configecho(void)
{
	if(echo < 0 || echo > 1) {
		printf("ECHO must be 0 or 1\n");
		exit(EXIT_FAILURE);
	}
	echo = !echo; /* Toggle echo setting (gets re-inverted and set later) */
}

static void checkcolour(void)
{
	if(usecolour < 0 || usecolour > 1) {
		printf("COLOUR must be 0 or 1\n");
		exit(EXIT_FAILURE);
	}
}

static void checkloadbaud(void)
{
	if(loadbaud < 2400 || loadbaud > 57600) {
		printf("LOADBAUD must be between 2400 and 57600 baud inclusive.\n");
		exit(EXIT_FAILURE);
	}
}

static void checkfgcol(void)
{
	if(fgcol < 0 || fgcol > 15) {
		printf("TEXTCOL must be between 0 and 15 inclusive.\n");
		exit(EXIT_FAILURE);
	}
}

static void checkbgcol(void)
{
	if(bgcol < 0 || bgcol > 7) {
		printf("BACKCOL must be between 0 and 7 inclusive.\n");
		exit(EXIT_FAILURE);
	}
}

static void readconfig(void)
{
	FILE *fp;
	char buf[40];
	char code;
  float f;

	fp = fopen("SLOWBP2.INI", "rt");
	if(fp == NULL) {
		printf("Can't open configuration file SLOWBP2.INI");
		exit(EXIT_FAILURE);
	}
	while(fgets(buf, 40, fp) != NULL) {
		strupr(buf);
		if(sscanf(buf, "PORT=%d", &base) == 1)
			setport();
		else if(sscanf(buf, "ID=\"%[^\"]", cwidmsg) == 1)
			;
		else if(sscanf(buf, "CODE=%c", &code) == 1)
			setcode(code);
		else if(sscanf(buf, "FREQ=%f", &f) == 1)
			tonefreq = txfreq = freq = f;
		else if(sscanf(buf, "RTSINVERT=%d", &rtsinvert) == 1)
			checkrtsinvert();
		else if(sscanf(buf, "DTRINVERT=%d", &dtrinvert) == 1)
			checkdtrinvert();
		else if(sscanf(buf, "ECHO=%d", &echo) == 1)
			configecho();
		else if(sscanf(buf, "COLOUR=%d", &usecolour) == 1)
			checkcolour();
		else if(sscanf(buf, "LOADBAUD=%d", &loadbaud) == 1)
			checkloadbaud();
		else if(sscanf(buf, "TEXTCOL=%d", &fgcol) == 1)
			checkfgcol();
		else if(sscanf(buf, "BACKCOL=%d", &bgcol) == 1)
			checkbgcol();
	}
	fclose(fp);
}

int main(void)
{
	int key, ret, chr;


	puts(version);
	base = combases[0]; /* Default to COM1 */
	vlcode = TRUE; /* Default to ASCII (vlcode gets inverted later) */
	readconfig();
	while(1) {
		ret = loaddsk("SLOWBP2.DSK", loadbaud);
		if(ret) {
			printf("\n\nFailed to load DSK: %s\n"
							"Press R to retry, any other key to abort.\n", loaderrs[ret]);
			if(toupper(getch()) == 'R')
				continue;
			else
				return 1;
		}
		else
			break;
	}
	delay(10); /* Ensure DSK program is ready */
	initcom(1556, 1); /* Reset COM port */
	init_tables(); /* Load varicode tables */
	setbuffernotify(0); /* Turn buffer empty notification off */
	setfreq(0); /* Set rx centre frequency */
	sendbyte('N'); /* Set tx centre frequency */
	dsp_buf_empty = FALSE;
	senddata(0x0008); /* Ensure tx is off */
	txon = FALSE;
	ptt_off();
	displayinit(); /* Initialise display */
	togglecode(); /* Invert vlcode and show status */
	toggle_echo(); /* Invert echo and show status */
	read_dsp_status(0x03); /* Show DSP buffer status */
	while(!quit) {
		if(inportb(base+5) & 0x01)
			getrxdata();
		if(rxbufinptr != rxbufoutptr)
			receive();
		if(file_send) {
			if(!dsp_buf_full) {
				chr = fgetc(snd_fp);
				if(chr == EOF)
					do_file_send(); /* Turn file send off */
				else
					sendchar(chr);
			}
		}
		if(!txon && dsp_buf_empty && ptton) {
			ptt_off();
			setbuffernotify(0);
		}
		if(kbhit()) {
			key = getch();
			switch(key) {
				case 27: /* Esc */
					quit = 1;
					break;
				case 0: /* Extended key */
					key = getch();
					switch(key) {
						case 72: /* Up arrow */
							setfreq(1);
							break;
						case 80: /* Down arrow */
							setfreq(-1);
							break;
						case 59: /* F1 */
							toggletxtrack();
							break;
						case 60: /* F2 */
							toggletx();
							break;
						case 61: /* F3 */
							cwident(cwidmsg);
							break;
						case 62: /* F4 */
							togglecode();
							break;
						case 63: /* F5 */
							togglesound();
							break;
						case 64: /* F6 */
							do_file_send();
							break;
						case 65: /* F7 */
							toggle_capture();
							break;
						case 66: /* F8 */
							toggle_echo();
							break;
						case 67: /* F9 */
							dcd = !dcd;
							showdcd();
							break;
					}
					break;
				default:
					if(!dsp_buf_full)
						sendchar(key);
			}
		}
	}
	restoredisplay(); /* Restore original display settings */
	return EXIT_SUCCESS;
}